/* addrecord.c

CGI program that saves posted record from a form into a text file.
This is for implementing a Rolodex-like textural database.

Looks at the following form entry names to constructs the path and get data:
   user
   catetory
   filename
   RecordToAdd

This needs to be SUID for the owner of data direcotry.  Otherwise,
the program runs as user "nobody", and won't be able to write to
data directory.

cp addrecord path_to_addrecord_executable
chmod u+s path_to_addrecord_executable
ls -l path_to_addrecord_executable
(confirm: -rwsr-xr-x and if owner is appropriate)

*/

#include "config.h"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define MAX_ENTRIES 10000



typedef struct {
    char *name;
    char *val;
} entry;

char *makeword(char *line, char stop);
char *fmakeword(FILE *f, char stop, int *len);
char x2c(char *what);
void unescape_url(char *url);
void plustospace(char *str);
int getpid(void);
void redo_form(char *message, char *user, char *content, char *filename);

void append_int(char *str, int inumber)
{
char tnumbuf[64];
	sprintf(tnumbuf, "%d", inumber);
	strcat(str, tnumbuf);
}	


static char *badchars = " !\"&'*,/:;<>?[\\]^`~|";
int isbadcharforfilename(int ch)
{
char *ptr;
int chbad;
int ibad = 0;
    ptr = badchars;
    while( (chbad = *ptr++) != '\0'){
	if(ch == chbad) {
	    ibad = 1;
	    break;
	}
    }
    return ibad;
}

/* Mainly convert filename so it won't have any spaces.
   Caller must allocate storage for *outfile. */

char *clean_filename(char *infile, char *outfile)
{
char *pi, *po;
int ch;
    pi = infile;
    po = outfile;
    while((ch = *pi++) != '\0') {
	if(isbadcharforfilename(ch) || ch < ' ' || ch > 'z')
	    *po++ = '_';
	else
	    *po++ = ch;
    }

    return outfile;
}


void main(int argc, char *argv[]) {
    entry entries[MAX_ENTRIES];
    int x,m=0;
    int cl, cl2, ch, empty=0;
    char outfile[1024], timestr[64], username[32], bufstr[1024];
    char *extension;
    char *user, *category, *content, *filename;
    char *pptr;
    FILE *fp;
    time_t timenow;

    time(&timenow);
    strcpy(timestr, ctime(&timenow));
    timestr[24] = '\0';

    printf("Content-type: text/html%c%c",10,10);

    if(strcmp(getenv("REQUEST_METHOD"),"POST")) {
        printf("This script should be referenced with a METHOD of POST.\n");
        printf("If you don't understand this, see this ");
        printf("<A HREF=\"http://www.ncsa.uiuc.edu/SDG/Software/Mosaic/Docs/fill-out-forms/overview.html\">forms overview</A>.%c",10);
        exit(1);
    }
    if(strcmp(getenv("CONTENT_TYPE"),"application/x-www-form-urlencoded")) {
        printf("This script can only be used to decode form results. \n");
        exit(1);
    }
    cl = cl2 = atoi(getenv("CONTENT_LENGTH"));

    for(x=0;cl && (!feof(stdin));x++) {
        m=x;
        entries[x].val = fmakeword(stdin,'&',&cl);
        plustospace(entries[x].val);
        unescape_url(entries[x].val);
        entries[x].name = makeword(entries[x].val,'=');
    }

    /* ### all fields have been parsed out into a structure now */

    /* fine various names */
    user = category = content = filename = NULL;

    for(x=0; x <= m; x++) {
	if((strcmp(entries[x].name, "user") == 0) && (strlen(entries[x].val) > 0)) {
	   user = entries[x].val;
	   strncpy(username, entries[x].val, 30);
	}
	if((strcmp(entries[x].name, "category") == 0) && (strlen(entries[x].val) > 0)) {
	   category = entries[x].val;
	}
	if((strcmp(entries[x].name, "filename") == 0) && (strlen(entries[x].val) > 0)) {
	   filename = entries[x].val;
	}
	if(strcmp(entries[x].name, "RecordToAdd") == 0) {
	    content = entries[x].val;
	    pptr = entries[x].val;
	    if(strlen(pptr) < 2)
		empty = 1;		/* Empty record */
	}
    } /* end of for() */

    /* Check if we accept or reject.  Empty content is allowed as a signal for reindex request.
       If there is content but any of other fields is missing, reject. */

    if(!empty && (!user || !category || !filename)) {
	redo_form("INCOMPLETE FORM: Please completely fill out all fields !!",
		user, content, filename);
	exit(0);
    }

    if(!user)
	   strcpy(username, "Nobody");

    /* ###### All fields OK now, process ! ################## */
    extension = ".txt";
    strcpy(outfile, BASE_DIR);		/* set the base path */
    if(category) {
	strcat(outfile, category);	/* add category to path */
	strcat(outfile, "/");
	if(strncmp(category, "Rolodex", 7) == 0)
	    extension = ".rolo";		/* special extension for rolodex */
    }
    else  /* just in case no category specified */
	strcat(outfile, "Incoming/");
    if(filename) {
        clean_filename(filename, bufstr);
        strcat(outfile, bufstr);		/* add cleaned filename stem to path */
    }
    else {	/* just in case no filename */
	   strcat(outfile, "NoName-");
	   append_int(outfile, getpid());
    }
    strcat(outfile, extension);		/* Add extention. File name complete */


    printf("<H2>Add Record Results</H2>\n<hr>\n");

    if(!empty) {
	fp = fopen(outfile, "a");
	if(fp != NULL) {
	    /* Add filename to path */
    
	    fprintf(fp, "%s\n\n", content);
	    fprintf(fp,"--\n##Date: %s\n##By: %s\n##Host: %s [%s]\n\n",
		    timestr, username, getenv("REMOTE_HOST"), getenv("REMOTE_ADDR"));
	    pptr = content;
	    while((ch = *pptr++) != '\0'){
		if(ch == '\n')
		    printf("<br>\n");
		else putchar(ch);
	    }
	    printf("<p><hr>\n");
	    fclose(fp);
	}
	else {
	    printf("<b>Add Record Failed:</b> Could not open file: %s\n", outfile);
	    printf("<p><hr>\n");
	}

    }  /* end if (!empty) */

    if(!empty)
	printf("Record has been written to file: <b>%s</b><br><p>\n", outfile);
    else
	printf("Empty record, but an <b>index update</b> has been queued.<br><p>\n");

    printf("##Date: %s<br>\n##By: %s<br>\n", timestr, username);
    printf("The new record will appear in the index within %s.<p>\n", INDEX_INTERVAL);

    /* We wrote a record. Indicate that index update is needed. */
    /* This file should be deleted by script that updates the index. */
    strcpy(outfile, BASE_DIR);		/* set the base path */
    strcat(outfile, INDEX_DIRTY);
    if((fp = fopen(outfile, "w")) != NULL) {
	fprintf(fp, "%s\n", timestr);
	fclose(fp);
    }


}
